from kili.utils import row_generator_from_paginated_calls
from test.utils import mocked_count_method, mocked_query_method

TEST_CASES = [
    {
        "case": "AAU, When I query objects with first=50, I get the first 50 objects",
        "args": {"first": 50},
        "expected_result": ({"id": i} for i in range(50)),
    },
    {
        "case": "AAU, When I query objects with first=200, I get the first 200 objects",
        "args": {"first": 200},
        "expected_result": ({"id": i} for i in range(200)),
    },
    {
        "case": "AAU, When I query objects with first=50 and skip=20, I skip the 20"
        " first objects I get the 50 objects following",
        "args": {"first": 50, "skip": 20},
        "expected_result": ({"id": i} for i in range(20, 70)),
    },
    {
        "case": "AAU, When I query objects with first=50 and skip=20, I skip the 20"
        "first objects I get the 200 following objects",
        "args": {"first": 200, "skip": 20},
        "expected_result": ({"id": i} for i in range(20, 220)),
    },
    {
        "case": "AAU, When I query objects with first=25100, I get the first 25100 objects",
        "args": {"first": 25100},
        "expected_result": ({"id": i} for i in range(25100)),
    },
    {
        "case": "AAU, When I query objects with first=20 and disable_tqdm, I get the first 2O objects",
        "args": {"first": 20, "disable_tqdm": True},
        "expected_result": ({"id": i} for i in range(20)),
    },
    {
        "case": "AAU, When I query objects with first=20, skip=20 and disable_tqdm, I  skip the 20"
        "first objects and I get the 20 following objects",
        "args": {"first": 20, "skip": 20, "disable_tqdm": True},
        "expected_result": ({"id": i} for i in range(20, 40)),
    },
    {
        "case": "AAU, When I query objects with first=0, I get no objects",
        "args": {"first": 0},
        "expected_result": iter(()),
    },
]


def test_row_generator_from_paginated_calls():
    """
        Simulates a count query result by returning a list of ids
    """
    for test_case in TEST_CASES:
        case_name = test_case["case"]
        expected = test_case["expected_result"]
        skip = test_case["args"].get('skip', 0)
        first = test_case["args"].get('first')
        disable_tqdm = test_case["args"].get('disable_tqdm', False)

        actual = row_generator_from_paginated_calls(
            skip,
            first,
            mocked_count_method,
            {},
            mocked_query_method,
            {first, skip},
            [],
            disable_tqdm)
        assert all(a == b for a, b in zip(actual, expected)), \
            f"Test case \"{case_name}\" failed"
        assert type(actual).__name__ == "generator", \
            f"Test case \"{case_name}\" failed"



"""
Asset mutations
"""

from typing import List, Optional, Union
from functools import partial

from typeguard import typechecked

from ...helpers import (Compatible,
                        convert_to_list_of_none,
                        format_metadata,
                        format_result,
                        is_none_or_empty)
from ...queries.project import QueriesProject
from .queries import (GQL_DELETE_MANY_FROM_DATASET,
                      GQL_UPDATE_PROPERTIES_IN_ASSETS)
from .helpers import process_append_many_to_dataset_parameters
from ...constants import NO_ACCESS_RIGHT
from ...orm import Asset


class MutationsAsset:
    """
    Set of Asset mutations
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @Compatible(['v1', 'v2'])
    @typechecked
    def append_many_to_dataset(
            self,
            project_id: str,
            content_array: Optional[List[str]] = None,
            external_id_array: Optional[List[str]] = None,
            is_honeypot_array: Optional[List[bool]] = None,
            status_array: Optional[List[str]] = None,
            json_content_array: Optional[List[List[Union[dict, str]]]] = None,
            json_metadata_array: Optional[List[dict]] = None):
        # pylint: disable=line-too-long
        """Append assets to a project.

        Args:
            project_id: Identifier of the project
            content_array: List of elements added to the assets of the project
                Must not be None except if you provide json_content_array.

                - For a `TEXT` project, the content can be either raw text, or URLs to TEXT assets.
                - For an `IMAGE` / `PDF` project, the content can be either URLs or paths to existing
                    images/pdf on your computer.
                - For a `VIDEO`  project, the content must be hosted on a web server,
                    and you point Kili to your data by giving the URLs.
            external_id_array: List of external ids given to identify the assets.
                If None, random identifiers are created.
            is_honeypot_array:  Whether to use the asset for honeypot
            status_array: By default, all imported assets are set to `TODO`. Other options:
                `ONGOING`, `LABELED`, `REVIEWED`.
            json_content_array: Useful for `FRAME` or `TEXT` projects only.

                - For `FRAME` projects, each element is a sequence of frames, i.e. a
                    list of URLs to images or a list of paths to images.
                - For `TEXT` projects, each element is a json_content dict,
                    formatted according to documentation [on how to import
                rich-text assets](https://github.com/kili-technology/kili-playground/blob/master/recipes/import_text_assets.ipynb)
            json_metadata_array: The metadata given to each asset should be stored in a json like dict with keys.

                - Add metadata visible on the asset with the following keys: `imageUrl`, `text`, `url`.
                    Example for one asset: `json_metadata_array = [{'imageUrl': '','text': '','url': ''}]`.
                - For video, you can specify a value with key 'processingParameters' to specify the sampling rate (default: 30).
                    Example for one asset: `json_metadata_array = [{'processingParameters': {'framesPlayedPerSecond': 10}}]`.

        Returns:
            A result object which indicates if the mutation was successful, or an error message.

        Examples:
            >>> kili.append_many_to_dataset(
                    project_id=project_id,
                    content_array=['https://upload.wikimedia.org/wikipedia/en/7/7d/Lenna_%28test_image%29.png'])

        !!! example "Recipe"
            - For more detailed examples on how to import assets,
                see [the recipe](https://github.com/kili-technology/kili-playground/blob/master/recipes/import_assets.ipynb)
                or [other examples](https://docs.kili-technology.com/recipes/importing-data) in our documentation.
            - For more detailed examples on how to import text assets,
                see [the recipe](https://github.com/kili-technology/kili-playground/blob/master/recipes/import_text_assets.ipynb).
        """
        playground = QueriesProject(self.auth)
        projects = playground.projects(project_id)
        assert len(projects) == 1, NO_ACCESS_RIGHT
        input_type = projects[0]['inputType']
        data, request = process_append_many_to_dataset_parameters(input_type,
                                                                  content_array,
                                                                  external_id_array,
                                                                  is_honeypot_array,
                                                                  status_array,
                                                                  json_content_array,
                                                                  json_metadata_array)
        variables = {
            'data': data,
            'where': {'id': project_id}
        }
        result = self.auth.client.execute(request, variables)
        return format_result('data', result, Asset)

    @Compatible(['v2'])
    @typechecked
    def update_properties_in_assets(self,
                                    asset_ids: List[str],
                                    external_ids: Optional[List[str]] = None,
                                    priorities: Optional[List[int]] = None,
                                    json_metadatas: Optional[List[Union[dict, str]]] = None,
                                    consensus_marks: Optional[List[float]] = None,
                                    honeypot_marks: Optional[List[float]] = None,
                                    to_be_labeled_by_array: Optional[List[List[str]]] = None,
                                    contents: Optional[List[str]] = None,
                                    json_contents: Optional[List[str]] = None,
                                    status_array: Optional[List[str]] = None,
                                    is_used_for_consensus_array: Optional[List[bool]] = None,
                                    is_honeypot_array: Optional[List[bool]] = None):
        """Update the properties of one or more assets.

        Args:
            asset_ids : The asset IDs to modify
            external_ids: Change the external id of the assets
            priorities : You can change the priority of the assets
                By default, all assets have a priority of 0.
            json_metadatas: The metadata given to an asset should be stored
                in a json like dict with keys `imageUrl`, `text`, `url`:
                `json_metadata = {'imageUrl': '','text': '','url': ''}`
            consensus_marks: Should be between 0 and 1
            honeypot_marks: Should be between 0 and 1
            to_be_labeled_by_array: If given, each element of the list should contain the emails of
                the labelers authorized to label the asset.
            contents: - For a NLP project, the content can be directly in text format
                - For an Image / Video / Pdf project, the content must be hosted on a web server,
                and you point Kili to your data by giving the URLs
            json_contents: - For a NLP project, the `json_content`
                is a text formatted using RichText
                - For a Video project, the`json_content` is a json containg urls pointing
                    to each frame of the video.
            status_array: Each element should be in `TODO`, `ONGOING`, `LABELED`, `REVIEWED`
            is_used_for_consensus_array: Whether to use the asset to compute consensus kpis or not
            is_honeypot_array: Whether to use the asset for honeypot

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_properties_in_assets(
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                    consensus_marks=[1, 0.7],
                    contents=[None, 'https://to/second/asset.png'],
                    external_ids=['external-id-of-your-choice-1', 'external-id-of-your-choice-2'],
                    honeypot_marks=[0.8, 0.5],
                    is_honeypot_array=[True, True],
                    is_used_for_consensus_array=[True, False],
                    priorities=[None, 2],
                    status_array=['LABELED', 'REVIEWED'],
                    to_be_labeled_by_array=[['test+pierre@kili-technology.com'], None],
            )
        """

        formatted_json_metadatas = None
        if json_metadatas is None:
            formatted_json_metadatas = None
        else:
            if isinstance(json_metadatas, list):
                formatted_json_metadatas = list(
                    map(format_metadata, json_metadatas))
            else:
                raise Exception('json_metadatas',
                                'Should be either a None or a list of None, string, list or dict')

        where_array = [{'id': asset_id} for asset_id in asset_ids]
        nb_assets_to_modify = len(where_array)
        if nb_assets_to_modify > 100:
            raise Exception(
                f'Too many assets ({nb_assets_to_modify}) updated at a time')
        data_array = [{} for i in range(len(where_array))]
        list_of_properties = [
            external_ids,
            priorities,
            formatted_json_metadatas,
            consensus_marks,
            honeypot_marks,
            to_be_labeled_by_array,
            contents,
            json_contents,
            status_array,
            is_used_for_consensus_array,
            is_honeypot_array
        ]
        data = list(map(partial(convert_to_list_of_none,
                                length=nb_assets_to_modify), list_of_properties))
        property_names = [
            'externalId',
            'priority',
            'jsonMetadata',
            'consensusMark',
            'honeypotMark',
            'toBeLabeledBy',
            'content',
            'jsonContent',
            'status',
            'isUsedForConsensus',
            'isHoneypot'
        ]
        to_be_labeled_by_array = data[5]
        should_reset_to_be_labeled_by_array = list(
            map(is_none_or_empty, to_be_labeled_by_array))
        for i, properties in enumerate(zip(*data)):
            for _property, property_value in zip(property_names, properties):
                data_array[i][_property] = property_value
        for i in range(nb_assets_to_modify):
            data_array[i]['shouldResetToBeLabeledBy'] = should_reset_to_be_labeled_by_array[i]
        variables = {
            'whereArray': where_array,
            'dataArray': data_array
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_ASSETS, variables)
        return format_result('data', result, Asset)

    @Compatible(['v1', 'v2'])
    @typechecked
    def delete_many_from_dataset(self, asset_ids: List[str]):
        """Delete assets from a project.

        Args:
            asset_ids: The list of identifiers of the assets to delete.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {'where': {'idIn': asset_ids}}
        result = self.auth.client.execute(
            GQL_DELETE_MANY_FROM_DATASET, variables)
        return format_result('data', result, Asset)



"""
Transfer Learning
"""

import subprocess
import time

from kili.client import Kili

SECONDS_TO_WAIT = 1
LABEL_FIELDS = ['isLatestLabelForUser',
                'labelType', 'jsonResponse', 'createdAt']
FIELDS = ['id', 'content', 'externalId'] + \
    [f'labels.{x}' for x in LABEL_FIELDS]


def get_labels_of_types(asset, label_types):
    """
    Extracts the latest labels from an asset

    Args:
        asset: the asset to extract the labels from
        label_types: type of label, either DEFAULT or REVIEW
    """
    labels = [label for label in asset['labels']
              if label['labelType'] in label_types and label['isLatestLabelForUser']]
    return sorted(labels, key=lambda label: label['createdAt'])


class TransferLearning:
    """
    TransferLearning
    """
    # pylint: disable=too-many-instance-attributes,too-many-arguments

    def __init__(
            self,
            api_key,
            api_endpoint,
            project_id,
            number_of_inferences,
            minimum_number_of_assets_to_launch_training=10):
        self.kili = Kili(api_key=api_key, api_endpoint=api_endpoint)
        self.api_key = api_key
        self.project_id = project_id
        self.current_inference_number = 0
        self.current_training_number = 0
        self.last_training_number = -1
        self.assets_seen_in_training = []
        self.minimum_number_of_assets_to_launch_training = \
            minimum_number_of_assets_to_launch_training
        self.number_of_inferences = number_of_inferences

    def get_assets_to_train(self):
        """
        Collects the assets to train on
        """
        assets = self.kili.assets(project_id=self.project_id, fields=FIELDS)
        assets_to_train = []
        for asset in assets:
            default_labels = get_labels_of_types(asset, ['DEFAULT'])
            review_labels = get_labels_of_types(asset, ['REVIEWED'])
            if len(review_labels) > 0:
                asset['labels'] = [review_labels[-1]]
                assets_to_train.append(asset)
            elif len(default_labels) == 1:
                asset['labels'] = [default_labels[-1]]
                assets_to_train.append(asset)
            elif len(review_labels) == 0 and len(default_labels) > 0:
                print(
                    f'Asset {asset["id"]} has several labels: it should be reviewed')
            else:
                continue
        return assets_to_train

    @staticmethod
    def train(assets_to_train):
        """
        Prints that training is ongoing
        """
        print(
            f'Launch training for {len(assets_to_train)} assets:'
            f' {[asset["id"] for asset in assets_to_train]}')

    def launch_train(self):
        """
        Launches the training
        """
        print('Launching train')
        time.sleep(SECONDS_TO_WAIT)
        assets_to_train = self.get_assets_to_train()
        if len(self.assets_seen_in_training) == 0:
            filtered_assets_to_train = assets_to_train
        else:
            filtered_assets_to_train = [
                asset for asset in assets_to_train
                if all(asset['id'] not in training for training in self.assets_seen_in_training)]
        if len(filtered_assets_to_train) >= self.minimum_number_of_assets_to_launch_training:
            print('Starting training')
            TransferLearning.train(filtered_assets_to_train)
            self.current_training_number += 1
            self.assets_seen_in_training.append(
                [asset['id'] for asset in filtered_assets_to_train])
        else:
            print('Not enough labeled assets to start training')

    def get_assets_to_predict(self):
        """
        Collects the assets to predict
        """
        assets = self.kili.assets(project_id=self.project_id, fields=FIELDS)
        assets_to_predict = []
        for asset in assets:
            labels = get_labels_of_types(asset, ['DEFAULT', 'REVIEWED'])
            if len(labels) == 0:
                assets_to_predict.append(asset)
        return assets_to_predict

    @staticmethod
    def predict(assets_to_predict):
        """
        Prints that prediction is ongoing
        """
        print(
            f'Launch inference for {len(assets_to_predict)} assets:'
            f' {[asset["id"] for asset in assets_to_predict]}')

    def launch_predict(self):
        """
        Launches the prediction
        """
        print('Launching prediction')
        time.sleep(SECONDS_TO_WAIT)
        if self.current_training_number == self.last_training_number:
            print('Inference will not be launched for now...')
            return
        assets_to_predict = self.get_assets_to_predict()
        if len(assets_to_predict) > 0:
            current_training_number = self.current_training_number
            TransferLearning.predict(assets_to_predict)
            self.last_training_number = current_training_number
            self.current_inference_number += 1

    @staticmethod
    def launch_tensorboard():
        """
        Launches the tensorboard
        """
        print('Starting Tensorboard...')
        with subprocess.Popen(['tensorboard', '--logdir=runs']) as proc:
            print(proc.stdout.read())

    def launch(self):
        """
        Launches the tensorboard, training and prediction
        """
        TransferLearning.launch_tensorboard()
        while self.current_inference_number < self.number_of_inferences:
            print(f'Iteration inference {self.current_inference_number}')
            self.launch_train()
            self.launch_predict()



# pylint: disable=invalid-name,too-many-instance-attributes

"""
GraphQL types
"""

from dataclasses import dataclass


@dataclass
class OrganizationWithoutUser:
    """
    A wrapper for Organization GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """
    id = 'id'
    address = 'address'
    canSeeDataset = 'canSeeDataset'
    city = 'city'
    country = 'country'
    license = 'license'
    name = 'name'
    numberOfAnnotations = 'numberOfAnnotations'
    numberOfLabeledAssets = 'numberOfLabeledAssets'
    numberOfHours = 'numberOfHours'
    zipCode = 'zipCode'


@dataclass
class UserWithoutProjectUsers:
    """
    A wrapper for User GraphQL object.
    """
    id = 'id'
    activated = 'activated'
    createdAt = 'createdAt'
    email = 'email'
    firstname = 'firstname'
    lastname = 'lastname'
    organization = OrganizationWithoutUser
    organizationId = 'organizationId'
    organizationIdCompute = 'organizationIdCompute'
    organizationRole = 'organizationRole'
    rights = 'rights'
    updatedAt = 'updatedAt'


@dataclass
class Organization(OrganizationWithoutUser):
    """
    A wrapper for Organization GraphQL object.
    """
    users = UserWithoutProjectUsers


@dataclass
class ProjectUserWithoutProject:
    """
    A wrapper for ProjectUser GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """
    id = 'id'
    activated = 'activated'
    consensusMark = 'consensusMark'
    consensusMarkPerCategory = 'consensusMarkPerCategory'
    consensusMarkCompute = 'consensusMarkCompute'
    honeypotMark = "honeypotMark"
    honeypotMarkCompute = 'honeypotMarkCompute'
    lastLabelingAt = "lastLabelingAt"
    lastLabelingAtCompute = 'lastLabelingAtCompute'
    numberOfAnnotations = "numberOfAnnotations"
    numberOfAnnotationsCompute = 'numberOfAnnotationsCompute'
    numberOfLabeledAssets = "numberOfLabeledAssets"
    numberOfLabels = "numberOfLabels"
    numberOfLabelsCompute = 'numberOfLabelsCompute'
    numberOfLabeledAssets = 'numberOfLabeledAssets'
    numberOfLabeledAssetsCompute = 'numberOfLabeledAssetsCompute'
    role = 'role'
    starred = 'starred'
    totalDuration = "totalDuration"
    totalDurationCompute = 'totalDurationCompute'
    user = UserWithoutProjectUsers


@dataclass
class ProjectWithoutDataset:
    """
    A wrapper for Project GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """
    id = 'id'
    assetMetadata = 'assetMetadata'
    assetMetadataCompute = 'assetMetadataCompute'
    author = UserWithoutProjectUsers
    consensusMark = 'consensusMark'
    consensusMarkPerCategory = 'consensusMarkPerCategory'
    consensusTotCoverage = 'consensusTotCoverage'
    createdAt = 'createdAt'
    description = 'description'
    honeypotMark = 'honeypotMark'
    inputType = 'inputType'
    instructions = 'instructions'
    interface = 'interface'
    interfaceCompute = 'interfaceCompute'
    jsonInterface = 'jsonInterface'
    minConsensusSize = 'minConsensusSize'
    mlTasks = 'mlTasks'
    mlTasksCompute = 'mlTasksCompute'
    numberOfRemainingAssets = 'numberOfRemainingAssets'
    numberOfAssets = 'numberOfAssets'
    numberOfAssetsWithSkippedLabels = 'numberOfAssetsWithSkippedLabels'
    numberOfOpenIssues = 'numberOfOpenIssues'
    numberOfSolvedIssues = 'numberOfSolvedIssues'
    numberOfReviewedAssets = 'numberOfReviewedAssets'
    readPermissionsForAssetsAndLabels = 'readPermissionsForAssetsAndLabels'
    reviewCoverage = 'reviewCoverage'
    rights = 'rights'
    roles = ProjectUserWithoutProject
    shouldRelaunchKpiComputation = 'shouldRelaunchKpiComputation'
    title = 'title'
    titleAndDescription = 'titleAndDescription'
    updatedAt = 'updatedAt'
    useHoneyPot = 'useHoneyPot'


@dataclass
class ProjectUser(ProjectUserWithoutProject):
    """
    A wrapper for ProjectUser GraphQL object.
    """
    project = ProjectWithoutDataset


@dataclass
class UserWithoutApiKey(UserWithoutProjectUsers):
    """
    A wrapper for User GraphQL object.
    """
    projectUsers = ProjectUser


@dataclass
class ApiKey:
    """
    A wrapper for ApiKey GraphQL object.
    """
    createdAt = 'createdAt'
    id = 'id'
    key = 'key'
    name = 'name'
    revoked = 'revoked'
    user = UserWithoutApiKey
    userId = 'userId'


@dataclass
class User(UserWithoutApiKey):
    """
    A wrapper for User GraphQL object.
    """
    apiKeys = ApiKey


@dataclass
class LabelWithoutLabelOf:
    """
    A wrapper for Label GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """
    id = 'id'
    assetIdCompute = 'assetIdCompute'
    authorIdCompute = 'authorIdCompute'
    author = User
    createdAt = 'createdAt'
    honeypotMark = 'honeypotMark'
    honeypotMarkCompute = 'honeypotMarkCompute'
    inferenceMark = 'inferenceMark'
    inferenceMarkCompute = 'inferenceMarkCompute'
    isLatestLabelForUser = 'isLatestLabelForUser'
    isLatestLabelForUserCompute = 'isLatestLabelForUserCompute'
    isLatestDefaultLabelForUser = 'isLatestDefaultLabelForUser'
    isLatestDefaultLabelForUserCompute = 'isLatestDefaultLabelForUserCompute'
    isLatestReviewLabelForUser = 'isLatestReviewLabelForUser'
    isLatestReviewLabelForUserCompute = 'isLatestReviewLabelForUserCompute'
    jsonResponse = 'jsonResponse'
    labelType = 'labelType'
    modelName = 'modelName'
    numberOfAnnotations = 'numberOfAnnotations'
    numberOfAnnotationsCompute = 'numberOfAnnotationsCompute'
    projectIdCompute = 'projectIdCompute'
    responseCompute = 'responseCompute'
    searchCompute = 'searchCompute'
    secondsToLabel = 'secondsToLabel'
    skipped = 'skipped'
    totalSecondsToLabel = 'totalSecondsToLabel'
    totalSecondsToLabelCompute = 'totalSecondsToLabelCompute'


@dataclass
class Lock:
    """
    A wrapper for Lock GraphQL object.
    """
    id = 'id'
    author = User
    authorIdCompute = 'authorIdCompute'
    createdAt = 'createdAt'
    lockType = 'lockType'
    lockOfIdCompute = 'lockOfIdCompute'


@dataclass
class CommentsWithoutCommentsOf:
    """
    A wrapper for Comment GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """
    id = 'id'
    author = ProjectUser
    createdAt = 'createdAt'
    text = 'text'
    updatedAt = 'updatedAt'


@dataclass
class Issue:
    """
    A wrapper for Issue GraphQL object.
    """
    id = 'id'
    assignee = ProjectUser
    author = ProjectUser
    comments = CommentsWithoutCommentsOf
    createdAt = 'createdAt'
    hasBeenSeen = 'hasBeenSeen'
    issueNumber = 'issueNumber'
    project = ProjectWithoutDataset
    status = 'status'
    type = 'type'
    updatedAt = 'updatedAt'


@dataclass
class Asset:
    """
    A wrapper for Asset GraphQL object.
    """
    id = 'id'
    consensusMark = 'consensusMark'
    consensusMarkCompute = 'consensusMarkCompute'
    consensusMarkPerCategory = 'consensusMarkPerCategory'
    content = 'content'
    contentJson = 'contentJson'
    contentJsonCompute = 'contentJsonCompute'
    createdAt = 'createdAt'
    duration = 'duration'
    durationCompute = 'durationCompute'
    externalId = 'externalId'
    honeypotMark = 'honeypotMark'
    honeypotMarkCompute = 'honeypotMarkCompute'
    inferenceMark = 'inferenceMark'
    inferenceMarkCompute = 'inferenceMarkCompute'
    isHoneypot = 'isHoneypot'
    isToBeLabeledBy = 'isToBeLabeledBy'
    issues = Issue
    isUsedForConsensus = 'isUsedForConsensus'
    jsonContent = 'jsonContent'
    jsonMetadata = 'jsonMetadata'
    labels = LabelWithoutLabelOf
    latestLabel = LabelWithoutLabelOf
    latestLabelCompute = LabelWithoutLabelOf
    locks = Lock
    metadataCompute = 'metadataCompute'
    metadata = 'metadata'
    numberOfValidLocks = 'numberOfValidLocks'
    numberOfValidLocksCompute = 'numberOfValidLocksCompute'
    ocrMetadata = 'ocrMetadata'
    priority = 'priority'
    project = ProjectWithoutDataset
    projectId = 'projectId'
    projectIdCompute = 'projectIdCompute'
    readPermissionsFromLabels = 'readPermissionsFromLabels'
    skippedCompute = 'skippedCompute'
    skipped = 'skipped'
    status = 'status'
    statusCompute = 'statusCompute'
    thumbnail = 'thumbnail'
    thumbnailCompute = 'thumbnailCompute'
    toBeLabeledBy = ProjectUser
    updatedAt = 'updatedAt'


@dataclass
class Label(LabelWithoutLabelOf):
    """
    A wrapper for Label GraphQL object.
    """
    labelOf = Asset


@dataclass
class Project(ProjectWithoutDataset):
    """
    A wrapper for Project GraphQL object.
    """
    dataset = Asset


@dataclass
class Notification:
    """
    A wrapper for Notification GraphQL object.
    """
    id = 'id'
    createdAt = 'createdAt'
    hasBeenSeen = 'hasBeenSeen'
    message = 'message'
    status = 'status'
    url = 'url'
    userID = 'userID'


@dataclass
class ProjectVersion:
    """
    A wrapper for ProjectVersion GraphQL object.
    """
    id = 'id'
    createdAt = 'createdAt'
    content = 'content'
    name = 'name'
    project = Project
    projectId = 'projectId'

# pylint: enable=invalid-name



"""
Utils
"""
from typing import List, Callable

import time
from tqdm import tqdm

# pylint: disable=too-many-arguments,too-many-locals


def row_generator_from_paginated_calls(
    skip: int,
    first: int,
    count_method: Callable[..., int],
    count_kwargs: dict,
    paged_call_method: Callable[..., List[dict]],
    paged_call_payload: dict,
    fields: List[str],
    disable_tqdm: bool,
):
    """
    Builds a row generator from paginated calls.

    Args:
        skip: Number of assets to skip (they are ordered by their date of creation, first to last).
        first: Maximum number of assets to return.
        count_method: Callable returning the number of available assets given `count_args`.
        count_kwargs: Keyword arguments passed to the `count_method`.
        paged_call_method: Callable returning the list of samples.
        paged_call_payload: Payload for the GraphQL query.
        fields: The list of strings to retrieved.
        disable_tqdm: If `True`, disables tqdm.
    """
    count_rows_retrieved = 0
    if not disable_tqdm:
        count_rows_available = count_method(**count_kwargs)
        count_rows_queried_total = min(count_rows_available,
                                       first) if first is not None else count_rows_available
    else:
        # dummy value that won't have any impact since tqdm is disabled
        count_rows_queried_total = 1 if first != 0 else 0
    count_rows_query_default = min(100, first or 100)
    throttling_delay = 60 / 250

    if count_rows_queried_total == 0:
        yield from ()
    else:
        with tqdm(total=count_rows_queried_total, disable=disable_tqdm) as pbar:
            while True:
                query_start = time.time()
                rows = paged_call_method(
                    count_rows_retrieved + skip,
                    count_rows_query_default,
                    paged_call_payload,
                    fields,
                )
                query_time = time.time() - query_start

                if query_time < throttling_delay:
                    time.sleep(throttling_delay - query_time)

                if rows is None or len(rows) == 0:
                    break

                for row in rows:
                    yield row

                count_rows_retrieved += len(rows)
                pbar.update(len(rows))
                if first is not None and count_rows_retrieved >= first:
                    break
